# Copyright (c) 2024 Zhendong Peng (pzd17@tsinghua.org.cn)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Literal, Optional

from wetext.utils import normalize


class Normalizer:
    def __init__(
        self,
        lang: Literal["auto", "en", "zh"] = "auto",
        operator: Literal["tn", "itn"] = "tn",
        traditional_to_simple: bool = False,
        full_to_half: bool = False,
        remove_interjections: bool = False,
        remove_puncts: bool = False,
        tag_oov: bool = False,
        enable_0_to_9: bool = False,
        remove_erhua: bool = False,
    ):
        """
        Initialize the Normalizer.

        Args:
            lang: The language of the text.
            operator: The operator to use.
            traditional_to_simple: Whether to convert traditional Chinese to simplified Chinese.
            full_to_half: Whether to convert full-width characters to half-width characters.
            remove_interjections: Whether to remove interjections.
            remove_puncts: Whether to remove punctuation.
            tag_oov: Whether to tag out-of-vocabulary words.
            remove_erhua: Whether to remove erhua for TN.
            enable_0_to_9: Whether to enable 0-to-9 conversion for ITN.
        """
        self.lang = lang
        self.operator = operator
        self.traditional_to_simple = traditional_to_simple
        self.full_to_half = full_to_half
        self.remove_interjections = remove_interjections
        self.remove_puncts = remove_puncts
        self.tag_oov = tag_oov
        self.enable_0_to_9 = enable_0_to_9
        self.remove_erhua = remove_erhua

    def normalize(
        self,
        text: str,
        lang: Optional[Literal["auto", "en", "zh"]] = None,
        operator: Optional[Literal["tn", "itn"]] = None,
        traditional_to_simple: bool = False,
        full_to_half: bool = False,
        remove_interjections: bool = False,
        remove_puncts: bool = False,
        tag_oov: bool = False,
        enable_0_to_9: bool = False,
        remove_erhua: bool = False,
    ) -> str:
        """
        Normalize the text.

        Args:
            text: The text to normalize.
            lang: The language of the text.
            operator: The operator to use.
            traditional_to_simple: Whether to convert traditional Chinese to simplified Chinese.
            full_to_half: Whether to convert full-width characters to half-width characters.
            remove_interjections: Whether to remove interjections.
            remove_puncts: Whether to remove punctuation.
            tag_oov: Whether to tag out-of-vocabulary words.
            remove_erhua: Whether to remove erhua for TN.
            enable_0_to_9: Whether to enable 0-to-9 conversion for ITN.
        Returns:
            The normalized text.
        """
        text = normalize(
            text,
            lang or self.lang,
            operator or self.operator,
            traditional_to_simple or self.traditional_to_simple,
            full_to_half or self.full_to_half,
            remove_interjections or self.remove_interjections,
            remove_puncts or self.remove_puncts,
            tag_oov or self.remove_puncts,
            enable_0_to_9 or self.enable_0_to_9,
            remove_erhua or self.remove_erhua,
        )
        return text



# Copyright (c) 2022 Zhendong Peng (pzd17@tsinghua.org.cn)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import string

EOS = "<EOS>"
TN_ORDERS = {
    "date": ["year", "month", "day"],
    "fraction": ["denominator", "numerator"],
    "measure": ["denominator", "numerator", "value"],
    "money": ["value", "currency"],
    "time": ["noon", "hour", "minute", "second"],
}
EN_TN_ORDERS = {
    "date": ["preserve_order", "text", "day", "month", "year"],
    "money": ["integer_part", "fractional_part", "quantity", "currency_maj"],
}
ITN_ORDERS = {
    "date": ["year", "month", "day"],
    "fraction": ["sign", "numerator", "denominator"],
    "measure": ["numerator", "denominator", "value"],
    "money": ["currency", "value", "decimal"],
    "time": ["hour", "minute", "second", "noon"],
}


class Token:

    def __init__(self, name):
        self.name = name
        self.order = []
        self.members = {}

    def append(self, key, value):
        self.order.append(key)
        self.members[key] = value

    def string(self, orders):
        output = self.name + " {"
        if self.name in orders.keys():
            if "preserve_order" not in self.members.keys() or self.members["preserve_order"] != "true":
                self.order = orders[self.name]

        for key in self.order:
            if key not in self.members.keys():
                continue
            output += ' {}: "{}"'.format(key, self.members[key])
        return output + " }"


class TokenParser:

    def __init__(self, lang, operator="tn"):
        assert lang in ("en", "zh")
        if lang == "en":
            if operator == "tn":
                self.orders = EN_TN_ORDERS
            else:
                raise NotImplementedError()
        else:
            if operator == "tn":
                self.orders = TN_ORDERS
            elif operator == "itn":
                self.orders = ITN_ORDERS

    def load(self, input):
        assert len(input) > 0
        self.index = 0
        self.text = input
        self.char = input[0]
        self.tokens = []

    def read(self):
        if self.index < len(self.text) - 1:
            self.index += 1
            self.char = self.text[self.index]
            return True
        self.char = EOS
        return False

    def parse_ws(self):
        not_eos = self.char != EOS
        while not_eos and self.char == " ":
            not_eos = self.read()
        return not_eos

    def parse_char(self, exp):
        if self.char == exp:
            self.read()
            return True
        return False

    def parse_chars(self, exp):
        ok = False
        for x in exp:
            ok |= self.parse_char(x)
        return ok

    def parse_key(self):
        assert self.char != EOS
        assert self.char not in string.whitespace

        key = ""
        while self.char in string.ascii_letters + "_":
            key += self.char
            self.read()
        return key

    def parse_value(self):
        assert self.char != EOS
        escape = False

        value = ""
        while self.char != '"':
            value += self.char
            escape = self.char == "\\"
            self.read()
            if escape:
                escape = False
                value += self.char
                self.read()
        return value

    def parse(self, input):
        self.load(input)
        while self.parse_ws():
            name = self.parse_key()
            self.parse_chars(" { ")

            token = Token(name)
            while self.parse_ws():
                if self.char == "}":
                    self.parse_char("}")
                    break
                key = self.parse_key()
                self.parse_chars(': "')
                value = self.parse_value()
                self.parse_char('"')
                token.append(key, value)
            self.tokens.append(token)

    def reorder(self, input):
        self.parse(input)
        output = ""
        for token in self.tokens:
            output += token.string(self.orders) + " "
        return output.strip()



# Copyright (c) 2025 Zhendong Peng (pzd17@tsinghua.org.cn)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from importlib.resources import files

from kaldifst import TextNormalizer as normalizer


def load_fst(fst_path) -> normalizer:
    """
    Load a FST from the fsts directory.

    Args:
        fst_path: The path to the FST file.
    Returns:
        The loaded FST.
    """
    fst_path = files("wetext.fsts").joinpath(fst_path)
    return normalizer(str(fst_path))


FSTS = {
    "preprocess": {
        "traditional_to_simple": load_fst("traditional_to_simple.fst"),
    },
    "en": {
        "tn": {
            "tagger": load_fst("en/tn/tagger.fst"),
            "verbalizer": load_fst("en/tn/verbalizer.fst"),
        }
    },
    "zh": {
        "tn": {
            "tagger": load_fst("zh/tn/tagger.fst"),
            "verbalizer": load_fst("zh/tn/verbalizer.fst"),
            "verbalizer_remove_erhua": load_fst("zh/tn/verbalizer_remove_erhua.fst"),
        },
        "itn": {
            "tagger": load_fst("zh/itn/tagger.fst"),
            "tagger_enable_0_to_9": load_fst("zh/itn/tagger_enable_0_to_9.fst"),
            "verbalizer": load_fst("zh/itn/verbalizer.fst"),
        },
    },
    "postprocess": {
        "full_to_half": load_fst("full_to_half.fst"),
        "remove_interjections": load_fst("remove_interjections.fst"),
        "remove_puncts": load_fst("remove_puncts.fst"),
        "tag_oov": load_fst("tag_oov.fst"),
    },
}



# Copyright (c) 2024 Zhendong Peng (pzd17@tsinghua.org.cn)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
from typing import Literal

from wetext.constants import FSTS
from wetext.token_parser import TokenParser


def get_lang(text: str) -> Literal["en", "zh"]:
    """
    Get the language of the text.

    Args:
        text: The text to get the language of.
    Returns:
        The language of the text.
    """
    contains_chinese = False
    for ch in text:
        if "\u4e00" <= ch <= "\u9fff":
            contains_chinese = True
    return "zh" if contains_chinese or text.isdigit() else "en"


def preprocess(text: str, traditional_to_simple: bool = False) -> str:
    """
    Preprocess the text before normalization.

    Args:
        text: The text to preprocess.
        traditional_to_simple: Whether to convert traditional Chinese to simplified Chinese.
    Returns:
        The preprocessed text.
    """
    if traditional_to_simple:
        text = FSTS["preprocess"]["traditional_to_simple"](text)
    return text.strip()


def postprocess(
    text: str,
    full_to_half: bool = False,
    remove_interjections: bool = False,
    remove_puncts: bool = False,
    tag_oov: bool = False,
) -> str:
    """
    Postprocess the text after normalization.

    Args:
        text: The text to postprocess.
        full_to_half: Whether to convert full-width characters to half-width characters.
        remove_interjections: Whether to remove interjections.
        remove_puncts: Whether to remove punctuation.
        tag_oov: Whether to tag out-of-vocabulary words.
    Returns:
        The postprocessed text.
    """
    if full_to_half:
        text = FSTS["postprocess"]["full_to_half"](text)
    if remove_interjections:
        text = FSTS["postprocess"]["remove_interjections"](text)
    if remove_puncts:
        text = FSTS["postprocess"]["remove_puncts"](text)
    if tag_oov:
        text = FSTS["postprocess"]["tag_oov"](text)
    return text.strip()


def should_normalize(text: str, operator: Literal["tn", "itn"], remove_erhua: bool = False) -> bool:
    """
    Check if the text should be normalized.

    Args:
        text: The text to check.
        operator: The operator to use.
        remove_erhua: Whether to remove erhua for TN.
    Returns:
        True if the text should be normalized, False otherwise.
    """
    if operator == "tn":
        if bool(re.search(r"\d", text)):
            return True
        if remove_erhua and re.search(r"儿|兒", text):
            return True
        return False
    return len(text) > 0


def reorder(text: str, lang: Literal["en", "zh"], operator: Literal["tn", "itn"]) -> str:
    """
    Reorder the text.

    Args:
        text: The text to reorder.
        lang: The language of the text.
    Returns:
        The reordered text.
    """
    return TokenParser(lang, operator).reorder(text)


def tag(text: str, lang: Literal["en", "zh"], operator: Literal["tn", "itn"], enable_0_to_9: bool = False) -> str:
    """
    Tag the text.

    Args:
        text: The text to tag.
        lang: The language of the text.
        operator: The operator to use.
        enable_0_to_9: Whether to enable 0-to-9 conversion for ITN.
    Returns:
        The tagged text.
    """
    tagger = FSTS[lang][operator]["tagger"]
    if enable_0_to_9 and lang == "zh" and operator == "itn":
        tagger = FSTS["zh"]["itn"]["tagger_enable_0_to_9"]
    return tagger(text).strip()


def verbalize(text: str, lang: Literal["en", "zh"], operator: Literal["tn", "itn"], remove_erhua: bool = False) -> str:
    """
    Verbalize the text.

    Args:
        text: The text to verbalize.
        lang: The language of the text.
        operator: The operator to use.
        remove_erhua: Whether to remove erhua for TN.
    Returns:
        The verbalized text.
    """
    verbalizer = FSTS[lang][operator]["verbalizer"]
    if remove_erhua and lang == "zh" and operator == "tn":
        verbalizer = FSTS["zh"]["tn"]["verbalizer_remove_erhua"]
    return verbalizer(text).strip()


def normalize(
    text: str,
    lang: Literal["auto", "en", "zh"] = "auto",
    operator: Literal["tn", "itn"] = "tn",
    traditional_to_simple: bool = False,
    full_to_half: bool = False,
    remove_interjections: bool = False,
    remove_puncts: bool = False,
    tag_oov: bool = False,
    enable_0_to_9: bool = False,
    remove_erhua: bool = False,
):
    """
    Normalize the text.

    Args:
        text: The text to normalize.
        lang: The language of the text.
        operator: The operator to use.
        traditional_to_simple: Whether to convert traditional Chinese to simplified Chinese.
        full_to_half: Whether to convert full-width characters to half-width characters.
        remove_interjections: Whether to remove interjections.
        remove_puncts: Whether to remove punctuation.
        tag_oov: Whether to tag out-of-vocabulary words.
        remove_erhua: Whether to remove erhua for TN.
        enable_0_to_9: Whether to enable 0-to-9 conversion for ITN.
    Returns:
        The normalized text.
    """
    text = preprocess(text, traditional_to_simple)
    if should_normalize(text, operator, remove_erhua):
        if lang == "auto":
            lang = get_lang(text)
        if lang == "en" and operator == "itn":
            # ITN for English is not supported now, using ITN for Chinese instead.
            lang = "zh"
        text = tag(text, lang, operator, enable_0_to_9)
        text = reorder(text, lang, operator)
        text = verbalize(text, lang, operator, remove_erhua)
    text = postprocess(text, full_to_half, remove_interjections, remove_puncts, tag_oov)
    return text



# Copyright (c) 2024 Zhendong Peng (pzd17@tsinghua.org.cn)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import click

from wetext import Normalizer


@click.command()
@click.argument("text")
@click.option("--lang", "-l", default="auto", type=click.Choice(["auto", "en", "zh"]))
@click.option("--operator", "-o", default="tn", type=click.Choice(["tn", "itn"]))
@click.option("--traditional-to-simple", is_flag=True, help="Convert traditional Chinese to simplified Chinese.")
@click.option("--full-to-half", is_flag=True, help="Convert full-width characters to half-width characters.")
@click.option("--remove-interjections", is_flag=True, help="Remove interjections.")
@click.option("--remove-puncts", is_flag=True, help="Remove punctuation.")
@click.option("--tag-oov", is_flag=True, help="Tag out-of-vocabulary words.")
@click.option("--enable-0-to-9", is_flag=True, help="Enable 0-to-9 conversion.")
@click.option("--remove-erhua", is_flag=True, help="Remove erhua.")
def main(
    text,
    lang,
    operator,
    traditional_to_simple,
    full_to_half,
    remove_interjections,
    remove_puncts,
    tag_oov,
    enable_0_to_9,
    remove_erhua,
):
    normalizer = Normalizer(
        lang,
        operator,
        traditional_to_simple,
        full_to_half,
        remove_interjections,
        remove_puncts,
        tag_oov,
        enable_0_to_9,
        remove_erhua,
    )
    text = normalizer.normalize(text)
    print(text)


if __name__ == "__main__":
    main()
